import os
from pathlib import Path
from typing import Iterator
import json
from logging import getLogger

from .task_queue import TaskQueue

logger = getLogger(__name__)


class Config:
    """A class for handling configuration settings for the aiomql package.

    Attributes:
        record_trades (bool): Whether to keep record of trades or not.
        filename (str): Name of the config file
        records_dir (str): Path to the directory where trade records are saved
        win_percentage (float): Percentage of achieved target profit in a trade to be considered a win
        login (int): Trading account number
        password (str): Trading account password
        server (str): Broker server
        path (str): Path to terminal file
        timeout (int): Timeout for terminal connection
        _initialize (bool): First time initialization flag
        state (dict): A global state dictionary for storing data across the framework
        root_dir (str): The root directory of the project
    Notes:
        By default, the config class looks for a file named aiomql.json.
        You can change this by passing the filename and/or the config_dir keyword argument(s) to the constructor
        or the load_config method.
        By passing reload=True to the load_config method, you can reload and search again for the config file.
    """

    login: int = 0
    password: str = ""
    server: str = ""
    path: str = ""
    timeout: int = 60000
    record_trades: bool = True
    filename: str = "aiomql.json"
    win_percentage: float = 0.85
    records_dir = Path(Path.home() / "Documents" / "Aiomql" / "Trade Records").mkdir(parents=True, exist_ok=True)
    config_dir: str = ''
    _initialize = True
    state: dict = {}
    root_dir: Path = Path('.').absolute().resolve()
    task_queue: TaskQueue = TaskQueue()

    _instance: 'Config'

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, **kwargs):
        reload = kwargs.pop('reload', False)
        [setattr(self, key, value) for key, value in kwargs.items()]
        self.load_config(reload=reload)

    def __setattr__(self, key, value):
        if key == 'root_dir':
            value = Path(value).absolute().resolve()
        if key == 'records_dir':
            value = self.create_records_dir(value)
        super().__setattr__(key, value)

    @staticmethod
    def walk_to_root(path: str | Path) -> Iterator[str]:
        if not os.path.exists(path):
            raise IOError("Starting path not found")

        if os.path.isfile(path):
            path = os.path.dirname(path)

        last_dir = None
        current_dir = os.path.abspath(path)
        while last_dir != current_dir:
            yield current_dir
            parent_dir = os.path.abspath(os.path.join(current_dir, os.path.pardir))
            last_dir, current_dir = current_dir, parent_dir

    def find_config(self):
        try:
            path = self.root_dir / self.config_dir
            for dirname in self.walk_to_root(path):
                check_path = os.path.join(dirname, self.filename)
                if os.path.isfile(check_path):
                    return check_path
            return None
        except Exception as _:
            return

    def create_records_dir(self, records_dir: str | Path):
        """Create records directory if it does not exist"""
        try:
            records_dir = Path(records_dir).absolute().resolve() if isinstance(records_dir, str) else records_dir
            records_dir.mkdir(parents=True, exist_ok=True)
            return records_dir
        except Exception as _:
            logger.warning("Unable to create records directory")

    def load_config(self, file: str = None, reload: bool = True, filename: str = None, config_dir: str = ''):
        """Load configuration settings from a file."""
        if not (self._initialize or reload):
            return
        self._initialize = False
        data = {}
        self.filename = filename or self.filename
        self.config_dir = config_dir or self.config_dir
        if (file := (file or self.find_config())) is None:
            logger.warning("No Config File Found")
        else:
            fh = open(file, mode="r")
            data = json.load(fh)
            fh.close()
        [setattr(self, key, value) for key, value in data.items()]

    def account_info(self) -> dict[str, int | str]:
        """Returns Account login details as found in the config object if available

        Returns:
            dict: A dictionary of login details
        """
        return {"login": self.login, "password": self.password, "server": self.server}


import asyncio
from typing import Coroutine, Callable, Awaitable
from logging import getLogger

logger = getLogger(__name__)


class QueueItem:
    def __init__(self, task: Callable | Awaitable | Coroutine, *args, **kwargs):
        self.task = task
        self.args = args
        self.kwargs = kwargs

    async def run(self):
        try:
            if asyncio.iscoroutinefunction(self.task):
                return await self.task(*self.args, **self.kwargs)
            else:
                return self.task(*self.args, **self.kwargs)
        except Exception as err:
            logger.error(f'Error in running {self.task.__name__} with {str(self.args)}, {self.kwargs}: {err}')


class TaskQueue:
    def __init__(self):
        self.queue = asyncio.Queue()

    def add(self, item: QueueItem):
        try:
            self.queue.put_nowait(item)
        except asyncio.QueueFull:
            return

    async def worker(self):
        while True:
            try:
                item: QueueItem = self.queue.get_nowait()
                await item.run()
                self.queue.task_done()
            except asyncio.QueueEmpty:
                return

    def add_task(self, item: Callable | Awaitable | Coroutine, *args, **kwargs):
        self.add(QueueItem(item, *args, **kwargs))
        asyncio.create_task(self.worker())

    async def start(self):
        await self.queue.join()


"""Sessions allow you to run code at specific times of the day."""
import asyncio
from datetime import time, timedelta, datetime
from asyncio import sleep, iscoroutinefunction
from typing import Literal, Callable
from logging import getLogger

from .positions import Positions

logger = getLogger(__name__)


def delta(obj: time):
    """Get the timedelta of a datetime.time object.

    Args:
        obj (datetime.time): A datetime.time object.
    """
    return timedelta(hours=obj.hour, minutes=obj.minute, seconds=obj.second, microseconds=obj.microsecond)


class Session:
    """A session is a time period between two datetime.time objects specified in utc.

    Attributes:
        start (datetime.time): The start time of the session.
        end (datetime.time): The end time of the session.
        on_start (str): The action to take when the session starts. Default is None.
        on_end (str): The action to take when the session ends. Default is None.
        custom_start (Callable): A custom function to call when the session starts. Default is None.
        custom_end (Callable): A custom function to call when the session ends. Default is None.
        name (str): A name for the session. Default is a combination of start and end.

    Methods:
        begin: Call the action specified in on_start or custom_start.
        close: Call the action specified in on_end or custom_end.
        action: Used by begin and close to call the action specified.
        delta: Get the timedelta of a datetime.time object.
        until: Get the seconds until the session starts from the current time.
    """
    def __init__(self, *, start: int | time, end: int | time,
                 on_start: Literal['close_all', 'close_win', 'close_loss', 'custom_start'] = None,
                 on_end: Literal['close_all', 'close_win', 'close_loss', 'custom_end'] = None,
                 custom_start: Callable = None, custom_end: Callable = None, name: str = ''):
        """Create a session.

        Keyword Args:
            start (int | datetime.time): The start time of the session in UTC.
            end (int | datetime.time): The end time of the session in UTC.
            on_start (Literal['close_all', 'close_win', 'close_loss', 'custom_start']): The action to take when the
                session starts. Default is None.
            on_end (Literal['close_all', 'close_win', 'close_loss', 'custom_end']): The action to take when the session
                ends. Default is None.
            custom_start (Callable): A custom function to call when the session starts. Default is None.
            custom_end (Callable): A custom function to call when the session ends. Default is None.
            name (str): A name for the session. Default is a combination of start and end.
        """
        self.start = start if isinstance(start, time) else time(hour=start)
        self.end = end if isinstance(end, time) else time(hour=end)
        self.on_start = on_start
        self.on_end = on_end
        self.custom_start = custom_start
        self.custom_end = custom_end
        self.name = name or f'{self.start} - {self.end}'

    def __contains__(self, item: time):
        if self.start > self.end:
            m1 = time(hour=23, minute=59, second=59, microsecond=9999)
            m2 = time(hour=0)
            return self.start <= item <= m1 or m2 <= item < self.end
        return self.start <= item < self.end

    def __str__(self):
        return f'{self.start}-->{self.name}-->{self.end}' if self.name else f'{self.start}-->{self.end}'

    def __repr__(self):
        return f'{self.start}-->{self.end}'

    def __len__(self):
        return (delta(self.start) - delta(self.end)).seconds

    async def begin(self):
        """Call the action specified in on_start or custom_start."""
        await self.action(self.on_start)

    async def close(self):
        """Call the action specified in on_end or custom_end."""
        await self.action(self.on_end)

    async def action(self, action):
        """Used by begin and close to call the action specified.

        Args:
            action (Literal['close_all', 'close_win', 'close_loss', 'custom_start', 'custom_end']): The action to take.
        """
        try:
            position = Positions()
            positions = await position.positions_get()

            match action:
                case 'close_all':
                    await asyncio.gather(*(position.close(price=pos.price_current, ticket=pos.ticket,
                                                          order_type=pos.type, volume=pos.volume,
                                                          symbol=pos.symbol) for pos in positions),
                                         return_exceptions=True)

                case 'close_win':
                    await asyncio.gather(
                        *(position.close(price=pos.price_current, ticket=pos.ticket, order_type=pos.type,
                                         volume=pos.volume, symbol=pos.symbol) for pos in positions if pos.profit > 0),
                        return_exceptions=True)

                case 'close_loss':
                    await asyncio.gather(
                        *(position.close(price=pos.price_current, ticket=pos.ticket, order_type=pos.type,
                                         volume=pos.volume, symbol=pos.symbol) for pos in positions if
                          pos.profit < 0), return_exceptions=True)

                case 'custom_end':
                    if iscoroutinefunction(self.custom_end):
                        await self.custom_end()
                    self.custom_end()

                case 'custom_start':
                    if iscoroutinefunction(self.custom_start):
                        await self.custom_start()
                    self.custom_start()

                case _:
                    pass
        except Exception as exe:
            logger.warning(f'Failed to call action {action} due to {exe}')

    def until(self):
        """Get the seconds until the session starts from the current time in seconds."""
        return (delta(self.start) - delta(datetime.utcnow().time())).seconds


class Sessions:
    """Sessions allow you to run code at specific times of the day. It is a collection of Session objects.
    Sessions are sorted by start time. The sessions object is an asynchronous context manager.

    Attributes:
        sessions (list[Session]): A list of Session objects.
        current_session (Session): The current session.

    Methods:
        find: Find a session that contains a datetime.time object.
        find_next: Find the next session that contains a datetime.time object.
        check: Check if the current session has started and if not, wait until it starts.
    """
    def __init__(self, *sessions: Session):
        self.sessions = list(sessions)
        self.sessions.sort(key=lambda x: (x.start, x.end))
        self.current_session = None

    def find(self, obj: time) -> Session | None:
        """Find a session that contains a datetime.time object.

        Args:
            obj (datetime.time): A datetime.time object.

        Returns:
            Session | None: A Session object or None if not found.
        """
        for session in self.sessions:
            if obj in session:
                return session
        return None

    def find_next(self, obj: time) -> Session:
        """Find the next session that contains a datetime.time object.

        Args:
            obj (datetime.time): A datetime.time object.

        Returns:
            Session: A Session object.
        """
        for session in self.sessions:
            if obj < session.start:
                return session
        return self.sessions[0]

    def __contains__(self, item: time):
        return True if self.find(item) is not None else False

    async def __aenter__(self):
        await self.check()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.current_session.close()

    async def check(self):
        """Check if the current session has started and if not, wait until it starts."""
        now = datetime.utcnow().time()
        current_session = self.find(now)
        if current_session:
            if self.current_session:
                if self.current_session == current_session:
                    return
                await self.current_session.close()

            self.current_session = current_session
            await self.current_session.begin()
            return

        await self.current_session.close() if self.current_session else ...
        current_session = self.find_next(now)
        secs = current_session.until() + 10
        logger.info(f'sleeping for {secs} seconds until next {current_session} session')
        await sleep(secs)
        self.current_session = current_session
        await self.current_session.begin()


from .meta_trader import MetaTrader
from .config import Config
from .models import *
from .constants import *
from .base import Base
from .errors import Error
from .exceptions import *
from .task_queue import TaskQueue


from ...symbol import Symbol
from ...core.exceptions import VolumeError

class ForexSymbol(Symbol):
    """Subclass of Symbol for Forex Symbols. Handles the conversion of currency and the computation of stop loss,
    take profit and volume.
    """
    def compute_points(self, *, amount: float, volume) -> float:
        """Compute the number of points required for a trade. Given the amount and the volume of the trade.
        Args:
            amount (float): Amount to trade
            volume (float): Volume to trade
        """
        points = amount / (volume * self.point * self.trade_contract_size)
        return points
    
    async def compute_volume_points(self, *, amount: float, points: float, use_limits=False, round_down: bool = True,
                                    adjust: float = False) -> tuple[float, float]:
        """Compute the volume and points required for a trade. Given the amount and the number of points.

        Args:
            amount (float): Amount to trade
            points (float): Number of points
            round_down: round down the computed volume to the nearest step default True
            adjust: Adjust the points if the computed volume is outside the range of permitted volumes
            use_limits: Adjust the computed volume to the nearest permitted volume if the computed volume is outside
        """
        amount = await self.check_amount(amount)
        volume = amount / (self.point * points * self.trade_contract_size)
        volume = self.round_off_volume(volume, round_down=round_down)
        if (chk_vol := self.check_volume(volume))[0]:
            if adjust:
                points = self.compute_points(amount=amount, volume=volume)
            return volume, points

        if use_limits:
            vol = chk_vol[1]
            if adjust:
                points = self.compute_points(amount=amount, volume=vol)
            return vol, points
        raise VolumeError(f"Incorrect Volume. Computed Volume outside the range of permitted volumes")

    async def compute_volume_sl(self, *, amount: float, price: float, sl: float,
                                use_limits=False, adjust: bool = False, round_down: bool = True) -> tuple[float, float]:
        amount = await self.check_amount(amount)
        volume = amount / ((price - sl) * self.trade_contract_size)
        volume = self.round_off_volume(volume, round_down=round_down)
        sign = volume / abs(volume) if volume else 1
        if (chk_vol := self.check_volume(abs(volume)))[0]:
            if adjust:
                sl = price - (amount / (volume * self.trade_contract_size))
                return abs(volume), sl
            return abs(volume), sl
        if use_limits:
            vol = chk_vol[1] * sign
            if adjust:
                sl = price - (amount / (vol * self.trade_contract_size))
            return abs(vol), sl
        raise VolumeError(f"Incorrect Volume. Computed Volume outside the range of permitted volumes")

    async def compute_volume(self, *, amount: float, points, use_limits=False, round_down=True) -> float:
        """Compute volume given an amount to risk and target points. Round the computed volume to the nearest step.

        Args:
            amount (float): Amount to risk. Given in terms of the account currency.
            points (float): Target points.
            use_limits (bool): If True, the computed volume checked against the maximum and minimum volume.
            round_down: round down the computed volume to the nearest step default True

        Returns:
            float: volume

        Raises:
            VolumeError: If the computed volume is less than the minimum volume or greater than the maximum volume.
        """
        amount = await self.check_amount(amount)
        volume = amount / (self.point * points * self.trade_contract_size)
        volume = self.round_off_volume(volume, round_down=round_down)
        if self.check_volume(volume)[0]:
            return volume
        if use_limits:
            return self.check_volume(volume)[1]
        raise VolumeError(f"Incorrect Volume. Computed Volume outside the range of permitted volumes")


"""Utility functions for aiomql."""

import decimal


def dict_to_string(data: dict, multi=True) -> str:
    """Convert a dict to a string. Useful for logging.

    Args:
        data (dict): The dict to convert.
        multi (bool, optional): If True, each key-value pair will be on a new line. Defaults to True.

    Returns:
        str: The string representation of the dict.
    """
    sep = '\n' if multi else ', '
    return f"{sep}".join(f"{key}: {value}" for key, value in data.items())


def round_off(value: float, step: float, round_down: bool = True) -> float:
    """Round off a number to the nearest step."""
    with decimal.localcontext() as ctx:
        ctx.rounding = decimal.ROUND_DOWN if round_down else decimal.ROUND_UP
        return float(decimal.Decimal(str(value)).quantize(decimal.Decimal(str(step))))


from enum import IntEnum, IntFlag

import MetaTrader5 as mt5

"""
MetaTrader5 constants as IntEnum types with Python style class names and nice string representation

Examples:
    >>> from aiomql import OrderFilling 
    >>> fok = OrderFilling.FOK
    >>> print(fok)
    "ORDER_FILLING_FOK"
"""


class Repr:
    __enum_name__ = ""

    def __repr__(self):
        return f"{self.__enum_name__}_{self.name}"


class TradeAction(Repr, IntEnum):
    """TRADE_REQUEST_ACTION Enum.
    
    Attributes:
        DEAL (int): Delete the pending order placed previously Place a trade order for an immediate execution with the
            specified parameters (market order).
        PENDING (int): Delete the pending order placed previously
        SLTP (int): Modify Stop Loss and Take Profit values of an opened position
        MODIFY (int): Modify the parameters of the order placed previously
        REMOVE (int): Delete the pending order placed previously
        CLOSE_BY (int): Close a position by an opposite one 
    """
    __enum_name__ = "TRADE_ACTION"
    DEAL = mt5.TRADE_ACTION_DEAL
    PENDING = mt5.TRADE_ACTION_PENDING
    SLTP = mt5.TRADE_ACTION_SLTP
    MODIFY = mt5.TRADE_ACTION_MODIFY
    REMOVE = mt5.TRADE_ACTION_MODIFY
    CLOSE_BY = mt5.TRADE_ACTION_CLOSE_BY


class OrderFilling(Repr, IntEnum):
    """ORDER_TYPE_FILLING Enum.
    
    Attributes:
        FOK (int): This execution policy means that an order can be executed only in the specified volume.
            If the necessary amount of a financial instrument is currently unavailable in the market, the order will
            not be executed. The desired volume can be made up of several available offers.

        IOC (int): An agreement to execute a deal at the maximum volume available in the market within the volume
            specified in the order. If the request cannot be filled completely, an order with the available volume will
            be executed, and the remaining volume will be canceled.

        RETURN (int): This policy is used only for market (ORDER_TYPE_BUY and ORDER_TYPE_SELL), limit and stop limit
            orders (ORDER_TYPE_BUY_LIMIT, ORDER_TYPE_SELL_LIMIT,ORDER_TYPE_BUY_STOP_LIMIT and
            ORDER_TYPE_SELL_STOP_LIMIT) and only for the symbols with Market or Exchange execution modes. If filled
            partially, a market or limit order with the remaining volume is not canceled, and is processed further.
            During activation of the ORDER_TYPE_BUY_STOP_LIMIT and ORDER_TYPE_SELL_STOP_LIMIT orders, an appropriate
            limit order ORDER_TYPE_BUY_LIMIT/ORDER_TYPE_SELL_LIMIT with the ORDER_FILLING_RETURN type is created.
    """
    __enum_name__ = "ORDER_FILLING"
    FOK = mt5.ORDER_FILLING_FOK
    IOC = mt5.ORDER_FILLING_IOC
    RETURN = mt5.ORDER_FILLING_RETURN


class OrderTime(Repr, IntEnum):
    """ORDER_TIME Enum.
    
    Attributes:
        GTC (int): Good till cancel order 
        DAY (int): Good till current trade day order 
        SPECIFIED (int): The order is active until the specified date 
        SPECIFIED_DAY (int): The order is active until 23:59:59 of the specified day. If this time appears to be out of
            a trading session, the expiration is processed at the nearest trading time.
    """
    __enum_name__ = "ORDER_TIME"
    GTC = mt5.ORDER_TIME_GTC
    DAY = mt5.ORDER_TIME_DAY
    SPECIFIED = mt5.ORDER_TIME_SPECIFIED
    SPECIFIED_DAY = mt5.ORDER_TIME_SPECIFIED_DAY


class OrderType(Repr, IntEnum):
    """ORDER_TYPE Enum.
    
    Attributes:
        BUY (int): Market buy order 
        SELL (int): Market sell order 
        BUY_LIMIT (int): Buy Limit pending order 
        SELL_LIMIT (int): Sell Limit pending order 
        BUY_STOP (int): Buy Stop pending order 
        SELL_STOP (int): Sell Stop pending order 
        BUY_STOP_LIMIT (int): Upon reaching the order price, Buy Limit pending order is placed at StopLimit price 
        SELL_STOP_LIMIT (int): Upon reaching the order price, Sell Limit pending order is placed at StopLimit price 
        CLOSE_BY (int): Order for closing a position by an opposite one 

    Properties:
        opposite (int): Gets the opposite of an order type
    """
    __enum_name__ = "ORDER_TYPE"
    BUY = mt5.ORDER_TYPE_BUY
    SELL = mt5.ORDER_TYPE_SELL
    BUY_LIMIT = mt5.ORDER_TYPE_BUY_LIMIT
    SELL_LIMIT = mt5.ORDER_TYPE_SELL_LIMIT
    BUY_STOP = mt5.ORDER_TYPE_BUY_STOP
    SELL_STOP = mt5.ORDER_TYPE_SELL_STOP
    BUY_STOP_LIMIT = mt5.ORDER_TYPE_BUY_STOP_LIMIT
    SELL_STOP_LIMIT = mt5.ORDER_TYPE_SELL_STOP_LIMIT
    CLOSE_BY = mt5.ORDER_TYPE_CLOSE_BY

    @property
    def opposite(self):
        """Gets the opposite of an order type for closing an open position

        Returns:
            int: integer value of opposite order type
        """
        return {0: 1, 1: 0, 2: 3, 3: 2, 4: 5, 5: 4, 6: 7, 7: 6, 8: 8}[self]


class BookType(Repr, IntEnum):
    """BOOK_TYPE Enum.

    Attributes:
        SELL (int): Sell order (Offer) 
        BUY (int): Buy order (Bid) 
        SELL_MARKET (int): Sell order by Market 
        BUY_MARKET (int): Buy order by Market 
    """
    __enum_name__ = "BOOK_TYPE"
    SELL = mt5.BOOK_TYPE_SELL
    BUY = mt5.BOOK_TYPE_BUY
    SELL_MARKET = mt5.BOOK_TYPE_SELL_MARKET
    BUY_MARKET = mt5.BOOK_TYPE_BUY_MARKET


class TimeFrame(Repr, IntEnum):
    """TIMEFRAME Enum.

    Attributes:
        M1 (int): One Minute
        M2 (int): Two Minutes
        M3 (int): Three Minutes
        M4 (int): Four Minutes
        M5 (int): Five Minutes
        M6 (int): Six Minutes
        M10 (int): Ten Minutes
        M15 (int): Fifteen Minutes
        M20 (int): Twenty Minutes
        M30 (int): Thirty Minutes
        H1 (int): One Hour
        H2 (int): Two Hours
        H3 (int): Three Hours
        H4 (int): Four Hours
        H6 (int): Six Hours
        H8 (int): Eight Hours
        D1 (int): One Day
        W1 (int): One Week
        MN1 (int): One Month

    Properties:
        time: return the value of the timeframe object in seconds. Used as a property

    Methods:
        get: get a timeframe object from a time value in seconds
    """
    __enum_name__ = "TIMEFRAME"

    def __str__(self):
        return self.name

    M1 = mt5.TIMEFRAME_M1
    M2 = mt5.TIMEFRAME_M2
    M3 = mt5.TIMEFRAME_M3
    M4 = mt5.TIMEFRAME_M4
    M5 = mt5.TIMEFRAME_M5
    M6 = mt5.TIMEFRAME_M6
    M10 = mt5.TIMEFRAME_M10
    M15 = mt5.TIMEFRAME_M15
    M20 = mt5.TIMEFRAME_M20
    M30 = mt5.TIMEFRAME_M30
    H1 = mt5.TIMEFRAME_H1
    H2 = mt5.TIMEFRAME_H2
    H3 = mt5.TIMEFRAME_H3
    H4 = mt5.TIMEFRAME_H4
    H6 = mt5.TIMEFRAME_H6
    H8 = mt5.TIMEFRAME_H8
    H12 = mt5.TIMEFRAME_H12
    D1 = mt5.TIMEFRAME_D1
    W1 = mt5.TIMEFRAME_W1
    MN1 = mt5.TIMEFRAME_MN1

    @property
    def time(self):
        """The number of seconds in a TIMEFRAME

        Returns:
            int: The number of seconds in a TIMEFRAME

        Examples:
            >>> t = TimeFrame.H1
            >>> print(t.time)
            3600
        """
        times = {1: 60, 2: 120, 3: 180, 4: 240, 5: 300, 6: 360, 10: 600, 15: 900, 20: 1200, 30: 1800, 16385: 3600,
                 16386: 7200, 16387: 10800, 16388: 14400, 16390: 21600, 16392: 28800, 16396: 43200, 16408: 86400,
                 32769: 604800, 49153: 2592000}
        return times[self]

    @classmethod
    def get(cls, time: int):
        times = {60: 1, 120: 2, 180: 3, 240: 4, 300: 5, 360: 6, 600: 10, 900: 15, 1200: 20, 1800: 30, 3600: 16385,
                 7200: 16386, 10800: 16387, 14400: 16388, 21600: 16390, 28800: 16392, 43200: 16396, 86400: 16408,
                 604800: 32769, 2592000: 49153}
        return TimeFrame(times[int(time)])


class CopyTicks(Repr, IntEnum):
    """COPY_TICKS Enum. This defines the types of ticks that can be requested using the copy_ticks_from() and
    copy_ticks_range() functions.

    Attributes:
        ALL (int): All ticks 
        INFO (int): Ticks containing Bid and/or Ask price changes 
        TRADE (int): Ticks containing Last and/or Volume price changes 
    """
    __enum_name__ = "COPY_TICKS"
    ALL = mt5.COPY_TICKS_ALL
    INFO = mt5.COPY_TICKS_INFO
    TRADE = mt5.COPY_TICKS_TRADE


class PositionType(Repr, IntEnum):
    """POSITION_TYPE Enum. Direction of an open position (buy or sell)

    Attributes:
        BUY (int): Buy 
        SELL (int): Sell
    """
    __enum_name__ = "POSITION_TYPE"
    BUY = mt5.POSITION_TYPE_BUY
    SELL = mt5.POSITION_TYPE_SELL


class PositionReason(Repr, IntEnum):
    """POSITION_REASON Enum. The reason for opening a position is contained in the POSITION_REASON Enum

    Attributes:
       CLIENT (int): The position was opened as a result of activation of an order placed from a desktop terminal
       MOBILE (int): The position was opened as a result of activation of an order placed from a mobile application
       WEB (int): The position was opened as a result of activation of an order placed from the web platform
       EXPERT (int): The position was opened as a result of activation of an order placed from an MQL5 program,
           i.e. an Expert Advisor or a script
    """
    __enum_name__ = "POSITION_REASON"
    CLIENT = mt5.POSITION_REASON_CLIENT
    MOBILE = mt5.POSITION_REASON_MOBILE
    WEB = mt5.POSITION_REASON_WEB
    EXPERT = mt5.POSITION_REASON_EXPERT


class DealType(Repr, IntEnum):
    """DEAL_TYPE enum. Each deal is characterized by a type, allowed values are enumerated in this enum

    Attributes:
        BUY (int): Buy
        SELL (int): Sell
        BALANCE (int): Balance
        CREDIT (int): Credit
        CHARGE (int): Additional Charge
        CORRECTION (int): Correction
        BONUS (int): Bonus
        COMMISSION (int): Additional Commission
        COMMISSION_DAILY (int): Daily Commission
        COMMISSION_MONTHLY (int): Monthly Commission
        COMMISSION_AGENT_DAILY (int): Daily Agent Commission
        COMMISSION_AGENT_MONTHLY (int): Monthly Agent Commission
        INTEREST (int): Interest Rate
        DEAL_DIVIDEND (int): Dividend Operations
        DEAL_DIVIDEND_FRANKED (int): Franked (non-taxable) dividend operations
        DEAL_TAX (int): Tax Charges

        BUY_CANCELED (int): Canceled buy deal. There can be a situation when a previously executed buy deal is canceled.
            In this case, the type of the previously executed deal (DEAL_TYPE_BUY) is changed to DEAL_TYPE_BUY_CANCELED,
            and its profit/loss is zeroized. Previously obtained profit/loss is charged/withdrawn using a separated
            balance operation

        SELL_CANCELED (int): Canceled sell deal. There can be a situation when a previously executed sell deal is
            canceled. In this case, the type of the previously executed deal (DEAL_TYPE_SELL) is changed to
            DEAL_TYPE_SELL_CANCELED, and its profit/loss is zeroized. Previously obtained profit/loss is
            charged/withdrawn using a separated balance operation.
    """
    __enum_name__ = "DEAL_TYPE"
    BUY = mt5.DEAL_TYPE_BUY
    SELL = mt5.DEAL_TYPE_SELL
    BALANCE = mt5.DEAL_TYPE_BALANCE
    CREDIT = mt5.DEAL_TYPE_CREDIT
    CHARGE = mt5.DEAL_TYPE_CHARGE
    CORRECTION = mt5.DEAL_TYPE_CORRECTION
    BONUS = mt5.DEAL_TYPE_BONUS
    COMMISSION = mt5.DEAL_TYPE_COMMISSION
    COMMISSION_DAILY = mt5.DEAL_TYPE_COMMISSION_DAILY
    COMMISSION_MONTHLY = mt5.DEAL_TYPE_COMMISSION_MONTHLY
    COMMISSION_AGENT_DAILY = mt5.DEAL_TYPE_COMMISSION_AGENT_DAILY
    COMMISSION_AGENT_MONTHLY = mt5.DEAL_TYPE_COMMISSION_AGENT_MONTHLY
    INTEREST = mt5.DEAL_TYPE_INTEREST
    BUY_CANCELED = mt5.DEAL_TYPE_BUY_CANCELED
    SELL_CANCELED = mt5.DEAL_TYPE_SELL_CANCELED
    DEAL_DIVIDEND = mt5.DEAL_DIVIDEND
    DEAL_DIVIDEND_FRANKED = mt5.DEAL_DIVIDEND_FRANKED
    DEAL_TAX = mt5.DEAL_TAX

    def __str__(self):
        if self.name in ('DEAL_DIVIDEND', 'DEAL_DIVIDEND_FRANKED', 'DEAL_TAX'):
            return self.name
        return super().__str__()


class DealEntry(Repr, IntEnum):
    """DEAL_ENTRY Enum. Deals differ not only in their types set in DEAL_TYPE enum, but also in the way they change
    positions. This can be a simple position opening, or accumulation of a previously opened position (market entering),
    position closing by an opposite deal of a corresponding volume (market exiting), or position reversing, if the
    opposite-direction deal covers the volume of the previously opened position.

    Attributes:
        IN (int): Entry In
        OUT (int): Entry Out
        INOUT (int): Reverse
        OUT_BY (int): Close a position by an opposite one
    """
    __enum_name__ = "DEAL_ENTRY"
    IN = mt5.DEAL_ENTRY_IN
    OUT = mt5.DEAL_ENTRY_OUT
    INOUT = mt5.DEAL_ENTRY_INOUT
    OUT_BY = mt5.DEAL_ENTRY_OUT_BY


class DealReason(Repr, IntEnum):
    """DEAL_REASON Enum. The reason for deal execution is contained in the DEAL_REASON property. A deal can be executed
    as a result of triggering of an order placed from a mobile application or an MQL5 program, as well as as a result
    of the StopOut event, variation margin calculation, etc.

    Attributes:
        CLIENT (int): The deal was executed as a result of activation of an order placed from a desktop terminal
        MOBILE (int): The deal was executed as a result of activation of an order placed from a desktop terminal
        WEB (int): The deal was executed as a result of activation of an order placed from the web platform
        EXPERT (int): The deal was executed as a result of activation of an order placed from an MQL5 program, i.e.
            an Expert Advisor or a script
        SL (int): The deal was executed as a result of Stop Loss activation
        TP (int): The deal was executed as a result of Take Profit activation
        SO (int): The deal was executed as a result of the Stop Out event
        ROLLOVER (int): The deal was executed due to a rollover
        VMARGIN (int): The deal was executed after charging the variation margin
        SPLIT (int): The deal was executed after the split (price reduction) of an instrument, which had an open
            position during split announcement
    """
    __enum_name__ = "DEAL_REASON"
    CLIENT = mt5.DEAL_REASON_CLIENT
    MOBILE = mt5.DEAL_REASON_MOBILE
    WEB = mt5.DEAL_REASON_WEB
    EXPERT = mt5.DEAL_REASON_EXPERT
    SL = mt5.DEAL_REASON_SL
    TP = mt5.DEAL_REASON_TP
    SO = mt5.DEAL_REASON_SO
    ROLLOVER = mt5.DEAL_REASON_ROLLOVER
    VMARGIN = mt5.DEAL_REASON_VMARGIN
    SPLIT = mt5.DEAL_REASON_SPLIT


class OrderReason(Repr, IntEnum):
    """ORDER_REASON Enum.

    Attributes:
        CLIENT (int): The order was placed from a desktop terminal
        MOBILE (int): The order was placed from a mobile application
        WEB (int): The order was placed from a web platform
        EXPERT (int): The order was placed from an MQL5-program, i.e. by an Expert Advisor or a script
        SL (int): The order was placed as a result of Stop Loss activation
        TP (int): The order was placed as a result of Take Profit activation
        SO (int): The order was placed as a result of the Stop Out event
    """
    __enum_name__ = "ORDER_REASON"
    CLIENT = mt5.ORDER_REASON_CLIENT
    MOBILE = mt5.ORDER_REASON_MOBILE
    WEB = mt5.ORDER_REASON_WEB
    EXPERT = mt5.ORDER_REASON_EXPERT
    SL = mt5.ORDER_REASON_SL
    TP = mt5.ORDER_REASON_TP
    SO = mt5.ORDER_REASON_SO


class SymbolChartMode(Repr, IntEnum):
    """SYMBOL_CHART_MODE Enum. A symbol price chart can be based on Bid or Last prices. The price selected for symbol
    charts also affects the generation and display of bars in the terminal.
    Possible values of the SYMBOL_CHART_MODE property are described in this enum

    Attributes:
        BID (int): Bars are based on Bid prices
        LAST (int): Bars are based on last prices
    """
    __enum_name__ = "SYMBOL_CHART_MODE"
    BID = mt5.SYMBOL_CHART_MODE_BID
    LAST = mt5.SYMBOL_CHART_MODE_LAST


class SymbolCalcMode(Repr, IntEnum):
    """SYMBOL_CALC_MODE Enum. The SYMBOL_CALC_MODE enumeration is used for obtaining information about how the margin
    requirements for a symbol are calculated.

    Attributes:
        FOREX (int): Forex mode - calculation of profit and margin for Forex
        FOREX_NO_LEVERAGE (int): Forex No Leverage mode – calculation of profit and margin for Forex symbols without
            taking into account the leverage
        FUTURES (int): Futures mode - calculation of margin and profit for futures
        CFD (int): CFD mode - calculation of margin and profit for CFD
        CFDINDEX (int): CFD index mode - calculation of margin and profit for CFD by indexes
        CFDLEVERAGE (int): CFD Leverage mode - calculation of margin and profit for CFD at leverage trading
        EXCH_STOCKS (int): Calculation of margin and profit for trading securities on a stock exchange
        EXCH_FUTURES (int): Calculation of margin and profit for trading futures contracts on a stock exchange
        EXCH_OPTIONS (int): value is 34
        EXCH_OPTIONS_MARGIN (int): value is 36
        EXCH_BONDS (int): Exchange Bonds mode – calculation of margin and profit for trading bonds on a stock exchange
        STOCKS_MOEX (int): Exchange MOEX Stocks mode –calculation of margin and profit for trading securities on MOEX
        EXCH_BONDS_MOEX (int): Exchange MOEX Bonds mode – calculation of margin and profit for trading bonds on MOEX

        SERV_COLLATERAL (int): Collateral mode - a symbol is used as a non-tradable asset on a trading account.
            The market value of an open position is calculated based on the volume, current market price, contract size
            and liquidity ratio. The value is included into Assets, which are added to Equity. Open positions of such
            symbols increase the Free Margin amount and are used as additional margin (collateral) for open positions
    """
    __enum_name__ = "SYMBOL_CALC_MODE"
    FOREX = mt5.SYMBOL_CALC_MODE_FOREX
    FOREX_NO_LEVERAGE = mt5.SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE
    FUTURES = mt5.SYMBOL_CALC_MODE_FUTURES
    CFD = mt5.SYMBOL_CALC_MODE_CFD
    CFDINDEX = mt5.SYMBOL_CALC_MODE_CFDINDEX
    CFDLEVERAGE = mt5.SYMBOL_CALC_MODE_CFDLEVERAGE
    EXCH_STOCKS = mt5.SYMBOL_CALC_MODE_EXCH_STOCKS
    EXCH_FUTURES = mt5.SYMBOL_CALC_MODE_EXCH_FUTURES
    EXCH_OPTIONS = mt5.SYMBOL_CALC_MODE_EXCH_OPTIONS
    EXCH_OPTIONS_MARGIN = mt5.SYMBOL_CALC_MODE_EXCH_OPTIONS_MARGIN
    EXCH_BONDS = mt5.SYMBOL_CALC_MODE_EXCH_BONDS
    EXCH_STOCKS_MOEX = mt5.SYMBOL_CALC_MODE_EXCH_STOCKS_MOEX
    EXCH_BONDS_MOEX = mt5.SYMBOL_CALC_MODE_EXCH_BONDS_MOEX
    SERV_COLLATERAL = mt5.SYMBOL_CALC_MODE_SERV_COLLATERAL


class SymbolTradeMode(Repr, IntEnum):
    """SYMBOL_TRADE_MODE Enum. There are several symbol trading modes. Information about trading modes of a certain
    symbol is reflected in the values this enumeration

    Attributes:
        DISABLED (int): Trade is disabled for the symbol
        LONGONLY (int): Allowed only long positions
        SHORTONLY (int): Allowed only short positions
        CLOSEONLY (int): Allowed only position close operations
        FULL (int): No trade restrictions
    """

    __enum_name__ = "SYMBOL_TRADE_MODE"
    DISABLED = mt5.SYMBOL_TRADE_MODE_DISABLED
    LONGONLY = mt5.SYMBOL_TRADE_MODE_LONGONLY
    SHORTONLY = mt5.SYMBOL_TRADE_MODE_SHORTONLY
    CLOSEONLY = mt5.SYMBOL_TRADE_MODE_CLOSEONLY
    FULL = mt5.SYMBOL_TRADE_MODE_FULL


class SymbolTradeExecution(Repr, IntEnum):
    """SYMBOL_TRADE_EXECUTION Enum. The modes, or execution policies, define the rules for cases when the price has
    changed or the requested volume cannot be completely fulfilled at the moment.

    Attributes:
        REQUEST (int): Executing a market order at the price previously received from the broker. Prices for a certain
            market order are requested from the broker before the order is sent. Upon receiving the prices, order
            execution at the given price can be either confirmed or rejected.

        INSTANT (int): Executing a market order at the specified price immediately. When sending a trade request to be
            executed, the platform automatically adds the current prices to the order.
            - If the broker accepts the price, the order is executed.
            - If the broker does not accept the requested price, a "Requote" is sent — the broker returns prices,
            at which this order can be executed.

        MARKET (int): A broker makes a decision about the order execution price without any additional discussion with the trader.
            Sending the order in such a mode means advance consent to its execution at this price.

        EXCHANGE (int): Trade operations are executed at the prices of the current market offers.
    """

    __enum_name__ = "SYMBOL_TRADE_EXECUTION"
    REQUEST = mt5.SYMBOL_TRADE_EXECUTION_REQUEST
    INSTANT = mt5.SYMBOL_TRADE_EXECUTION_INSTANT
    MARKET = mt5.SYMBOL_TRADE_EXECUTION_MARKET
    EXCHANGE = mt5.SYMBOL_TRADE_EXECUTION_EXCHANGE


class SymbolSwapMode(Repr, IntEnum):
    """SYMBOL_SWAP_MODE Enum. Methods of swap calculation at position transfer are specified in enumeration
    ENUM_SYMBOL_SWAP_MODE. The method of swap calculation determines the units of measure of the SYMBOL_SWAP_LONG and
    SYMBOL_SWAP_SHORT parameters. For example, if swaps are charged in the client deposit currency, then the values of
    those parameters are specified as an amount of money in the client deposit currency.

    Attributes:
        DISABLED (int): Swaps disabled (no swaps)
        POINTS (int): Swaps are charged in points
        CURRENCY_SYMBOL (int): Swaps are charged in money in base currency of the symbol
        CURRENCY_MARGIN (int): Swaps are charged in money in margin currency of the symbol
        CURRENCY_DEPOSIT (int): Swaps are charged in money, in client deposit currency

        INTEREST_CURRENT (int): Swaps are charged as the specified annual interest from the instrument price at
            calculation of swap (standard bank year is 360 days)

        INTEREST_OPEN (int): Swaps are charged as the specified annual interest from the open price of position
            (standard bank year is 360 days)

        REOPEN_CURRENT (int): Swaps are charged by reopening positions. At the end of a trading day the position is
            closed. Next day it is reopened by the close price +/- specified number of points
            (parameters SYMBOL_SWAP_LONG and SYMBOL_SWAP_SHORT)

        REOPEN_BID (int): Swaps are charged by reopening positions. At the end of a trading day the position is closed.
            Next day it is reopened by the current Bid price +/- specified number of
            points (parameters SYMBOL_SWAP_LONG and SYMBOL_SWAP_SHORT)
    """
    __enum_name__ = "SYMBOL_SWAP_MODE"
    DISABLED = mt5.SYMBOL_SWAP_MODE_DISABLED
    POINTS = mt5.SYMBOL_SWAP_MODE_POINTS
    CURRENCY_SYMBOL = mt5.SYMBOL_SWAP_MODE_CURRENCY_SYMBOL
    CURRENCY_MARGIN = mt5.SYMBOL_SWAP_MODE_CURRENCY_MARGIN
    CURRENCY_DEPOSIT = mt5.SYMBOL_SWAP_MODE_CURRENCY_DEPOSIT
    INTEREST_CURRENT = mt5.SYMBOL_SWAP_MODE_INTEREST_CURRENT
    INTEREST_OPEN = mt5.SYMBOL_SWAP_MODE_INTEREST_OPEN
    REOPEN_CURRENT = mt5.SYMBOL_SWAP_MODE_REOPEN_CURRENT
    REOPEN_BID = mt5.SYMBOL_SWAP_MODE_REOPEN_BID


class DayOfWeek(Repr, IntEnum):
    """DAY_OF_WEEK Enum.

    Attributes:
        SUNDAY (int): Sunday
        MONDAY (int): Monday
        TUESDAY (int): Tuesday
        WEDNESDAY (int): Wednesday
        THURSDAY (int): Thursday
        FRIDAY (int): Friday
        SATURDAY (int): Saturday
    """
    __enum__name__ = "DAY_OF_WEEK"
    SUNDAY = mt5.DAY_OF_WEEK_SUNDAY
    MONDAY = mt5.DAY_OF_WEEK_MONDAY
    TUESDAY = mt5.DAY_OF_WEEK_TUESDAY
    WEDNESDAY = mt5.DAY_OF_WEEK_WEDNESDAY
    THURSDAY = mt5.DAY_OF_WEEK_THURSDAY
    FRIDAY = mt5.DAY_OF_WEEK_FRIDAY
    SATURDAY = mt5.DAY_OF_WEEK_SATURDAY


class SymbolOrderGTCMode(Repr, IntEnum):
    """SYMBOL_ORDER_GTC_MODE Enum. If the SYMBOL_EXPIRATION_MODE property is set to SYMBOL_EXPIRATION_GTC
        (good till canceled), the expiration of pending orders, as well as of
        Stop Loss/Take Profit orders should be additionally set using the ENUM_SYMBOL_ORDER_GTC_MODE enumeration.

    Attributes:
        GTC (int): Pending orders and Stop Loss/Take Profit levels are valid for an unlimited period
            until theirConstants, Enumerations and explicit cancellation

        DAILY (int): Orders are valid during one trading day. At the end of the day, all Stop Loss and
            Take Profit levels, as well as pending orders are deleted.

        DAILY_NO_STOPS (int): When a trade day changes, only pending orders are deleted,
            while Stop Loss and Take Profit levels are preserved
    """
    __enum_name__ = "SYMBOL_ORDERS"
    GTC = mt5.SYMBOL_ORDERS_GTC
    DAILY = mt5.SYMBOL_ORDERS_DAILY
    DAILY_NO_STOPS = mt5.SYMBOL_ORDERS_DAILY_NO_STOPS


class SymbolOptionRight(Repr, IntEnum):
    """SYMBOL_OPTION_RIGHT Enum. An option is a contract, which gives the right, but not the obligation,
    to buy or sell an underlying asset (goods, stocks, futures, etc.) at a specified price on or before a specific date.
    The following enumerations describe option properties, including the option type and the right arising from it.

    Attributes:
        CALL (int): A call option gives you the right to buy an asset at a specified price.
        PUT (int): A put option gives you the right to sell an asset at a specified price.
    """
    __enum_name__ = "SYMBOL_OPTION_RIGHT"
    CALL = mt5.SYMBOL_OPTION_RIGHT_CALL
    PUT = mt5.SYMBOL_OPTION_RIGHT_PUT


class SymbolOptionMode(Repr, IntEnum):
    """SYMBOL_OPTION_MODE Enum.

    Attributes:
        EUROPEAN (int): European option may only be exercised on a specified date (expiration, execution date, delivery date)
        AMERICAN (int): American option may be exercised on any trading day or before expiry. The period within which
        a buyer can exercise the option is specified for it.
    """
    __enum_name__ = "SYMBOL_OPTION_MODE"
    EUROPEAN = mt5.SYMBOL_OPTION_MODE_EUROPEAN
    AMERICAN = mt5.SYMBOL_OPTION_MODE_AMERICAN


class AccountTradeMode(Repr, IntEnum):
    """ACCOUNT_TRADE_MODE Enum. There are several types of accounts that can be opened on a trade server.
    The type of account on which an MQL5 program is running can be found out using
    the ENUM_ACCOUNT_TRADE_MODE enumeration.

    Attributes:
        DEMO: Demo account
        CONTEST: Contest account
        REAL: Real Account
    """
    __enum_name__ = "ACCOUNT_TRADE_MODE"
    DEMO = mt5.ACCOUNT_TRADE_MODE_DEMO
    CONTEST = mt5.ACCOUNT_TRADE_MODE_CONTEST
    REAL = mt5.ACCOUNT_TRADE_MODE_REAL


class TickFlag(Repr, IntFlag):
    """TICK_FLAG Enum. TICK_FLAG defines possible flags for ticks. These flags are used to describe ticks obtained by the
    copy_ticks_from() and copy_ticks_range() functions.

    Attributes:
        BID (int): Bid price changed
        ASK (int): Ask price changed
        LAST (int): Last price changed
        VOLUME (int): Volume changed
        BUY (int): last Buy price changed
        SELL (int): last Sell price changed
        """
    __enum_name__ = "TICK_FLAG"
    BID = mt5.TICK_FLAG_BID
    ASK = mt5.TICK_FLAG_ASK
    LAST = mt5.TICK_FLAG_LAST
    VOLUME = mt5.TICK_FLAG_VOLUME
    BUY = mt5.TICK_FLAG_BUY
    SELL = mt5.TICK_FLAG_SELL


class TradeRetcode(Repr, IntEnum):
    """TRADE_RETCODE Enum. Return codes for order send/check operations

    Attributes:
        REQUOTE (int): Requote
        REJECT (int): Request rejected
        CANCEL (int): Request canceled by trader
        PLACED (int): Order placed
        DONE (int): Request completed
        DONE_PARTIAL (int): Only part of the request was completed
        ERROR (int): Request processing error
        TIMEOUT (int): Request canceled by timeout
        INVALID (int): Invalid request
        INVALID_VOLUME (int): Invalid volume in the request
        INVALID_PRICE (int): Invalid price in the request
        INVALID_STOPS (int): Invalid stops in the request
        TRADE_DISABLED (int): Trade is disabled
        MARKET_CLOSED (int): Market is closed
        NO_MONEY (int): There is not enough money to complete the request
        PRICE_CHANGED (int): Prices changed
        PRICE_OFF (int): There are no quotes to process the request
        INVALID_EXPIRATION (int): Invalid order expiration date in the request
        ORDER_CHANGED (int): Order state changed
        TOO_MANY_REQUESTS (int): Too frequent requests
        NO_CHANGES (int): No changes in request
        SERVER_DISABLES_AT (int): Autotrading disabled by server
        CLIENT_DISABLES_AT (int): Autotrading disabled by client terminal
        LOCKED (int): Request locked for processing
        FROZEN (int): Order or position frozen
        INVALID_FILL (int): Invalid order filling type
        CONNECTION (int): No connection with the trade server
        ONLY_REAL (int): Operation is allowed only for live accounts
        LIMIT_ORDERS (int): The number of pending orders has reached the limit
        LIMIT_VOLUME (int): The volume of orders and positions for the symbol has reached the limit
        INVALID_ORDER (int): Incorrect or prohibited order type
        POSITION_CLOSED (int): Position with the specified POSITION_IDENTIFIER has already been closed
        INVALID_CLOSE_VOLUME (int): A close volume exceeds the current position volume

        CLOSE_ORDER_EXIST (int): A close order already exists for a specified position. This may happen when working in
            the hedging system:
            · when attempting to close a position with an opposite one, while close orders for the position already exist
            · when attempting to fully or partially close a position if the total volume of the already present close
                orders and the newly placed one exceeds the current position volume

        LIMIT_POSITIONS (int): The number of open positions simultaneously present on an account can be limited by the
            server settings.After a limit is reached, the server returns the TRADE_RETCODE_LIMIT_POSITIONS error when
            attempting to place an order. The limitation operates differently depending on the position accounting type:
            · Netting — number of open positions is considered. When a limit is reached, the platform does not let
                placing new orders whose execution may increase the number of open positions. In fact, the platform
                allows placing orders only for the symbols that already have open positions.
                The current pending orders are not considered since their execution may lead to changes in the current
                positions but it cannot increase their number.

            · Hedging — pending orders are considered together with open positions, since a pending order activation
                always leads to opening a new position. When a limit is reached, the platform does not allow placing
                both new market orders for opening positions and pending orders.

        REJECT_CANCEL (int): The pending order activation request is rejected, the order is canceled.
        LONG_ONLY (int): The request is rejected, because the "Only long positions are allowed" rule is set for the
            symbol (POSITION_TYPE_BUY)
        SHORT_ONLY (int): The request is rejected, because the "Only short positions are allowed" rule is set for the
            symbol (POSITION_TYPE_SELL)
        CLOSE_ONLY (int): The request is rejected, because the "Only position closing is allowed" rule is set for the
            symbol
        FIFO_CLOSE (int): The request is rejected, because "Position closing is allowed only by FIFO rule" flag is set
            for the trading account (ACCOUNT_FIFO_CLOSE=true)
    """
    __enum_name__ = "TRADE_RETCODE"
    REQUOTE = mt5.TRADE_RETCODE_REQUOTE
    REJECT = mt5.TRADE_RETCODE_REJECT
    CANCEL = mt5.TRADE_RETCODE_CANCEL
    PLACED = mt5.TRADE_RETCODE_PLACED
    DONE = mt5.TRADE_RETCODE_DONE
    DONE_PARTIAL = mt5.TRADE_RETCODE_DONE_PARTIAL
    ERROR = mt5.TRADE_RETCODE_ERROR
    TIMEOUT = mt5.TRADE_RETCODE_TIMEOUT
    INVALID = mt5.TRADE_RETCODE_INVALID
    INVALID_VOLUME = mt5.TRADE_RETCODE_INVALID_VOLUME
    INVALID_PRICE = mt5.TRADE_RETCODE_INVALID_PRICE
    INVALID_STOPS = mt5.TRADE_RETCODE_INVALID_STOPS
    TRADE_DISABLED = mt5.TRADE_RETCODE_TRADE_DISABLED
    MARKET_CLOSED = mt5.TRADE_RETCODE_MARKET_CLOSED
    NO_MONEY = mt5.TRADE_RETCODE_NO_MONEY 
    PRICE_CHANGED = mt5.TRADE_RETCODE_PRICE_CHANGED
    PRICE_OFF = mt5.TRADE_RETCODE_PRICE_OFF
    INVALID_EXPIRATION = mt5.TRADE_RETCODE_INVALID_EXPIRATION
    ORDER_CHANGED = mt5.TRADE_RETCODE_ORDER_CHANGED
    TOO_MANY_REQUESTS = mt5.TRADE_RETCODE_TOO_MANY_REQUESTS
    NO_CHANGES = mt5.TRADE_RETCODE_NO_CHANGES
    SERVER_DISABLES_AT = mt5.TRADE_RETCODE_SERVER_DISABLES_AT
    CLIENT_DISABLES_AT = mt5.TRADE_RETCODE_CLIENT_DISABLES_AT
    LOCKED = mt5.TRADE_RETCODE_LOCKED
    FROZEN = mt5.TRADE_RETCODE_FROZEN
    INVALID_FILL = mt5.TRADE_RETCODE_INVALID_FILL
    CONNECTION = mt5.TRADE_RETCODE_CONNECTION
    ONLY_REAL = mt5.TRADE_RETCODE_ONLY_REAL
    LIMIT_ORDERS = mt5.TRADE_RETCODE_LIMIT_ORDERS
    LIMIT_VOLUME = mt5.TRADE_RETCODE_LIMIT_VOLUME
    INVALID_ORDER = mt5.TRADE_RETCODE_INVALID_ORDER
    POSITION_CLOSED = mt5.TRADE_RETCODE_POSITION_CLOSED
    INVALID_CLOSE_VOLUME = mt5.TRADE_RETCODE_INVALID_CLOSE_VOLUME
    CLOSE_ORDER_EXIST = mt5.TRADE_RETCODE_CLOSE_ORDER_EXIST
    LIMIT_POSITIONS = mt5.TRADE_RETCODE_LIMIT_POSITIONS
    REJECT_CANCEL = mt5.TRADE_RETCODE_REJECT_CANCEL
    LONG_ONLY = mt5.TRADE_RETCODE_LONG_ONLY
    SHORT_ONLY = mt5.TRADE_RETCODE_SHORT_ONLY
    CLOSE_ONLY = mt5.TRADE_RETCODE_CLOSE_ONLY
    FIFO_CLOSE = mt5.TRADE_RETCODE_FIFO_CLOSE
    

class AccountStopOutMode(Repr, IntEnum):
    """ACCOUNT_STOPOUT_MODE Enum.

    Attributes:
        PERCENT (int): Account stop out mode in percents
        MONEY (int): Account stop out mode in money
    """

    __enum_name__ = "ACCOUNT_STOPOUT_MODE"
    PERCENT = mt5.ACCOUNT_STOPOUT_MODE_PERCENT
    MONEY = mt5.ACCOUNT_STOPOUT_MODE_MONEY


class AccountMarginMode(Repr, IntEnum):
    """ACCOUNT_MARGIN_MODE Enum.

    Attributes:
        RETAIL_NETTING (int): Used for the OTC markets to interpret positions in the "netting"
            mode (only one position can exist for one symbol). The margin is calculated based on the symbol
            type (SYMBOL_TRADE_CALC_MODE).

        EXCHANGE (int): Used for the exchange markets. Margin is calculated based on the discounts specified in
            symbol settings. Discounts are set by the broker, but not less than the values set by the exchange.

        HEDGING (int): Used for the exchange markets where individual positions are possible
            (hedging, multiple positions can exist for one symbol). The margin is calculated based on the symbol
            type (SYMBOL_TRADE_CALC_MODE) taking into account the hedged margin (SYMBOL_MARGIN_HEDGED).
    """
    __enum_name__ = "ACCOUNT_MARGIN_MODE"
    RETAIL_NETTING = mt5.ACCOUNT_MARGIN_MODE_RETAIL_NETTING
    EXCHANGE = mt5.ACCOUNT_MARGIN_MODE_EXCHANGE
    RETAIL_HEDGING = mt5.ACCOUNT_MARGIN_MODE_RETAIL_HEDGING


import asyncio
import logging

from .tracker import Tracker
from ..traders import SimpleTrader
from ...symbol import Symbol
from ...trader import Trader
from ...candle import Candles
from ...strategy import Strategy
from ...core import TimeFrame, OrderType
from ...sessions import Sessions

logger = logging.getLogger(__name__)


class FingerTrap(Strategy):
    ttf: TimeFrame
    etf: TimeFrame
    trend: int
    fast_ema: int
    slow_ema: int
    entry_ema: int
    parameters: dict
    ecc: int
    tcc: int
    trader: Trader
    tracker: Tracker
    parameters = {"trend": 3, "fast_ema": 8, "slow_ema": 20, "etf": TimeFrame.M5,
                  "ttf": TimeFrame.H1, "entry_ema": 5, "tcc": 50, "ecc": 600}

    def __init__(self, *, symbol: Symbol, params: dict | None = None, trader: Trader = None, sessions: Sessions = None,
                 name: str = 'FingerTrap'):
        super().__init__(symbol=symbol, params=params, sessions=sessions, name=name)
        self.trader = trader or SimpleTrader(symbol=self.symbol)
        self.tracker: Tracker = Tracker(snooze=self.ttf.time)

    async def check_trend(self):
        try:
            candles: Candles = await self.symbol.copy_rates_from_pos(timeframe=self.ttf, count=self.tcc)
            if not ((current := candles[-1].time) >= self.tracker.trend_time):
                self.tracker.update(new=False, order_type=None)
                return
            self.tracker.update(new=True, trend_time=current)
            candles.ta.ema(length=self.slow_ema, append=True, fillna=0)
            candles.ta.ema(length=self.fast_ema, append=True, fillna=0)
            candles.rename(inplace=True, **{f"EMA_{self.fast_ema}": "fast", f"EMA_{self.slow_ema}": "slow"})

            fas = candles.ta_lib.above(candles.fast, candles.slow) # fast above slow
            fbs = candles.ta_lib.below(candles.fast, candles.slow) # fast below slow
            caf = candles.ta_lib.above(candles.close, candles.fast) # close above fast
            cbf = candles.ta_lib.below(candles.close, candles.fast) # close below fast
            current = candles[-2]
            if fas.iloc[-1] and caf.iloc[-1] and current.is_bullish():
                self.tracker.update(trend="bullish")

            elif fbs.iloc[-1] and cbf.iloc[-1] and current.is_bearish():
                self.tracker.update(trend="bearish")
            else:
                self.tracker.update(trend="ranging", snooze=self.ttf.time, order_type=None)
        except Exception as err:
            logger.error(f"{err} for {self.symbol} in {self.__class__.__name__}.check_trend")
            self.tracker.update(snooze=self.ttf.time, order_type=None)

    async def confirm_trend(self):
        try:
            candles = await self.symbol.copy_rates_from_pos(timeframe=self.etf, count=self.ecc)
            if not ((current := candles[-1].time) >= self.tracker.entry_time):
                self.tracker.update(new=False, order_type=None)
                return

            self.tracker.update(new=True, entry_time=current)
            candles.ta.ema(length=self.entry_ema, append=True)
            candles.rename(**{f"EMA_{self.entry_ema}": "ema"})
            cae = candles.ta_lib.cross(candles.close, candles.ema)
            cbe = candles.ta_lib.cross(candles.close, candles.ema, above=False)
            if self.tracker.bullish and any([cae.iloc[-1], cae.iloc[-2]]):
                self.tracker.update(snooze=self.ttf.time, order_type=OrderType.BUY)
            elif self.tracker.bearish and any([cbe.iloc[-1], cbe.iloc[-2]]):
                self.tracker.update(snooze=self.ttf.time, order_type=OrderType.SELL)
            else:
                self.tracker.update(snooze=self.etf.time, order_type=None)
        except Exception as err:
            logger.error(f"{err} for {self.symbol} in {self.__class__.__name__}.confirm_trend\n")
            self.tracker.update(snooze=self.etf.time, order_type=None)

    async def watch_market(self):
        await self.check_trend()
        if not self.tracker.ranging:
            await self.confirm_trend()

    async def trade(self):
        logger.info(f"Trading {self.symbol}")
        async with self.sessions as sess:
            await self.sleep(self.ttf.time)
            while True:
                await sess.check()
                try:
                    await self.watch_market()
                    if not self.tracker.new:
                        await asyncio.sleep(2)
                        continue
                    if self.tracker.order_type is None:
                        await self.sleep(self.tracker.snooze)
                        continue
                    await self.trader.place_trade(order_type=self.tracker.order_type, parameters=self.parameters)
                    await self.sleep(self.tracker.snooze)
                except Exception as err:
                    logger.error(f"{err} For {self.symbol} in {self.__class__.__name__}.trade\n")
                    await self.sleep(self.ttf.time)


import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Sequence, Coroutine, Callable

from .strategy import Strategy


class Executor:
    """Executor class for running multiple strategies on multiple symbols concurrently.

    Attributes:
        executor (ThreadPoolExecutor): The executor object.
        workers (list): List of strategies.
        coroutines (dict[Coroutine, dict]): A dictionary of coroutines to run in the executor
        functions (dict[Callable, dict]): A dictionary of functions to run in the executor
    """

    def __init__(self, bot=None):
        self.executor = ThreadPoolExecutor
        self.workers: list[type(Strategy)] = []
        self.coroutines: dict[Coroutine | Callable: dict] = {}
        self.functions: dict[Callable: dict] = {}
        self.bot: 'Bot' = bot

    def add_function(self, func: Callable, kwargs: dict):
        self.functions[func] = kwargs | {'bot': self.bot}

    def add_coroutine(self, coro: Coroutine, kwargs: dict):
        self.coroutines[coro] = kwargs | {'bot': self.bot}

    def add_workers(self, strategies: Sequence[type(Strategy)]):
        """Add multiple strategies at once

        Args:
            strategies (Sequence[Strategy]): A sequence of strategies.
        """
        self.workers.extend(strategies)

    def remove_workers(self):
        """Removes any worker running on a symbol not successfully initialized."""
        self.workers = [worker for worker in self.workers if worker.symbol in self.bot.symbols]

    def add_worker(self, strategy: type(Strategy)):
        """Add a strategy instance to the list of workers

        Args:
            strategy (Strategy): A strategy object
        """
        self.workers.append(strategy)

    @staticmethod
    def trade(strategy: type(Strategy)):
        """Wraps the coroutine trade method of each strategy with 'asyncio.run'.

        Args:
            strategy (Strategy): A strategy object
        """
        asyncio.run(strategy.trade())

    def run(self, func, kwargs: dict):
        """
        Run a coroutine function

        Args:
            func: The coroutine. A variadic function.
            kwargs: A dictionary of keyword arguments for the function
        """
        asyncio.run(func(**kwargs))

    async def execute(self, workers: int = 0):
        """Run the strategies with a threadpool executor.

        Args:
            workers: Number of workers to use in executor pool. Defaults to zero which uses all workers.

        Notes:
            No matter the number specified, the executor will always use a minimum of 5 workers.
        """
        workers = workers or sum([len(self.workers), len(self.functions), len(self.coroutines)])
        workers = max(workers, 5)
        loop = asyncio.get_running_loop()
        with self.executor(max_workers=workers) as executor:
            [loop.run_in_executor(executor, self.trade, worker) for worker in self.workers]
            [loop.run_in_executor(executor, self.run, coro, kwargs) for coro, kwargs in self.coroutines.items()]
            [loop.run_in_executor(executor, func, kwargs) for func, kwargs in self.functions.items()]


"""Risk Assessment and Management"""
from .account import Account
from .positions import Positions


class RAM:
    account: Account
    risk_to_reward: float
    risk: float
    points: float
    pips: float
    min_amount: float
    max_amount: float
    balance_level: float = 50
    loss_limit: int = 3

    def __init__(self, *, risk_to_reward: float = 1, risk: float = 0.01, **kwargs):
        """Initialize Risk Assessment and Management with the provided keyword arguments.

        Keyword Args:
            risk_to_reward (float): Risk to reward ratio. Defaults to 1
            risk (float): Percentage of account balance to risk per trade 0.01 # 1%
            kwargs: extra keyword arguments are set as object attributes
        """
        self.risk_to_reward = risk_to_reward
        self.risk = risk
        self.account = Account()
        [setattr(self, key, value) for key, value in kwargs.items()]

    async def get_amount(self) -> float:
        """Calculate the amount to risk per trade as a percentage of balance.

        Returns:
            float: Amount to risk per trade
        """
        await self.account.refresh()
        return self.account.balance * self.risk

    async def check_losing_positions(self) -> bool:
        positions = await Positions().positions_get()
        positions.sort(key=lambda pos: pos.time_msc)
        loosing = [trade for trade in positions if trade.profit <= 0]
        return len(loosing) > self.loss_limit

    async def check_balance_level(self) -> bool:
        await self.account.refresh()
        balance_level = (self.account.margin / self.account.balance) * 100
        return balance_level >= self.balance_level


from logging import getLogger

from ..symbols import ForexSymbol
from ...ram import RAM
from ...core.models import OrderType
from ...trader import Trader

logger = getLogger(__name__)


class SimpleTrader(Trader):
    """A simple trader class"""
    def __init__(self, *, symbol: ForexSymbol, ram: RAM = None):
        """Initializes the order object and RAM instance

        Args:
            symbol (Symbol): Financial instrument
            ram (RAM): Risk Assessment and Management instance
        """
        ram = ram or RAM(risk_to_reward=2)
        super().__init__(symbol=symbol, ram=ram)

    async def create_order(self, *, order_type: OrderType):
        """Complete the order object with the required values. Creates a simple order.

        Args:
            order_type (OrderType): Type of order
        """
        losing = await self.ram.check_losing_positions()
        if losing:
            raise RuntimeError(f"More than {self.ram.loss_limit} losing positions")
        amount = await self.ram.get_amount()
        points = self.symbol.compute_points(amount=amount, volume=self.symbol.volume_min)
        self.order.volume = self.symbol.volume_min
        self.order.type = order_type
        self.order.comment = self.parameters.get('name', 'SimpleTrader')
        tick = await self.symbol.info_tick()
        self.set_trade_stop_levels(points=points, tick=tick)

    async def place_trade(self, order_type: OrderType, parameters: dict = None):
        """Places a trade based on the order_type."""
        try:
            self.parameters |= parameters or {}
            await self.create_order(order_type=order_type)
            if not await self.check_order():
                return
            await self.send_order()
        except Exception as err:
            logger.error(f"{err} in {self.__class__.__name__}.place_trade for {self.symbol.name}")


from logging import getLogger

from .core.models import AccountInfo, SymbolInfo
from .core.exceptions import LoginError

logger = getLogger(__name__)


class Account(AccountInfo):
    """A class for managing a trading account. A singleton class.
    A subclass of AccountInfo. All AccountInfo attributes are available in this class.

    Attributes:
        connected (bool): Status of connection to MetaTrader 5 Terminal
        symbols (set[SymbolInfo]): A set of available symbols for the financial market.

    Notes:
        Other Account properties are defined in the AccountInfo class.
    """
    _instance: 'Account'
    connected: bool
    symbols = set()
    
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.login:
            acc = self.config.account_info()
            self.set_attributes(**acc)

    async def refresh(self):
        """Refreshes the account instance with the latest account details from the MetaTrader 5 terminal"""
        account_info = await self.mt5.account_info()
        acc = account_info._asdict()
        self.set_attributes(**acc)

    async def __aenter__(self) -> 'Account':
        """Connect to a trading account and return the account instance.
        Async context manager for the Account class.

        Returns:
            Account: An instance of the Account class

        Raises:
            LoginError: If login fails
        """
        res = await self.sign_in()
        if not res:
            raise LoginError('Login failed')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.mt5.shutdown()
        self.connected = False

    async def sign_in(self) -> bool:
        """Connect to a trading account.

        Returns:
            bool: True if login was successful else False
        """
        acc = self.get_dict(include={'login', 'server', 'password'})
        self.connected = await self._login(acc=acc)
        if self.connected:
            await self.refresh()
            self.symbols = await self.symbols_get()
            return self.connected
        await self.mt5.shutdown()
        return False

    async def _login(self, *, acc:dict, tries=3):
        res = False
        if tries == 0:
            return False
        ini = await self.mt5.initialize(**acc, path=self.config.path)
        if ini:
            res = await self.mt5.login(**acc)
        if ini and res:
            return True
        else:
            return await self._login(acc=acc, tries=tries-1)

    def has_symbol(self, symbol: str | SymbolInfo):
        """Checks to see if a symbol is available for a trading account.

        Args:
            symbol (str | SymbolInfo):

        Returns:
            bool: True if symbol is present otherwise False
        """
        try:
            return str(symbol) in {s.name for s in self.symbols}
        except Exception as err:
            logger.warning(f'Error: {err}; {symbol} not available in this market')
            return False

    async def symbols_get(self) -> set[SymbolInfo]:
        """Get all financial instruments from the MetaTrader 5 terminal available for the current account.

        Returns:
            set[Symbol]: A set of available symbols.
        """
        syms = await self.mt5.symbols_get()
        return {SymbolInfo(name=sym.name) for sym in syms}


"""Order Class"""
from logging import getLogger

from .core.models import TradeRequest, OrderSendResult, OrderCheckResult, TradeOrder, SymbolInfo
from .core.constants import TradeAction, OrderTime, OrderFilling
from .core.exceptions import SymbolError, OrderError
from .symbol import Symbol
logger = getLogger(__name__)


class Order(TradeRequest):
    """Trade order related functions and properties. Subclass of TradeRequest."""

    def __init__(self, **kwargs):
        """Initialize the order object with keyword arguments, symbol must be provided.
        Provide default values for action, type_time and type_filling if not provided.

        Args:
            **kwargs: Keyword arguments must match the attributes of TradeRequest as well as the attributes of
             Order class as specified in the annotations in the class definition.

        Default Values:
            action (TradeAction.DEAL): Trade action
            type_time (OrderTime.DAY): Order time
            type_filling (OrderFilling.FOK): Order filling
        """
        if 'symbol' in kwargs:
            kwargs['symbol'] = str(kwargs['symbol'])
        self.action = kwargs.pop('action', TradeAction.DEAL)
        self.type_time = kwargs.pop('type_time', OrderTime.DAY)
        self.type_filling = kwargs.pop('type_filling', OrderFilling.FOK)
        super().__init__(**kwargs)

    async def orders_total(self):
        """Get the number of active orders.

        Returns:
            (int): total number of active orders
        """
        return await self.mt5.orders_total()

    async def orders(self) -> tuple[TradeOrder]:
        """Get the list of active orders for the current symbol.

        Returns:
            tuple[TradeOrder]: A Tuple of active trade orders as TradeOrder objects
        """
        orders = await self.mt5.orders_get(symbol=self.symbol)
        if orders is None:
            raise OrderError(f'Failed to get orders for {self.symbol} due to {self.mt5.error.description}')
        orders = (TradeOrder(**order._asdict()) for order in orders)
        return tuple(orders)

    async def check(self) -> OrderCheckResult:
        """Check funds sufficiency for performing a required trading operation and the possibility to execute it at

        Returns:
            OrderCheckResult: An OrderCheckResult object

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_check(self.dict)
        if res is None:
            raise OrderError(f'Failed to check order due to {self.mt5.error.description}')
        return OrderCheckResult(**res._asdict())

    async def send(self) -> OrderSendResult:
        """Send a request to perform a trading operation from the terminal to the trade server.

        Returns:
             OrderSendResult: An OrderSendResult object

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_send(self.dict)
        if res is None:
            raise OrderError(f'Failed to send order {self.symbol} due to {self.mt5.error.description}')
        return OrderSendResult(**res._asdict())

    async def calc_margin(self) -> float:
        """Return the required margin in the account currency to perform a specified trading operation.

        Returns:
            float: Returns float value if successful

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_calc_margin(self.type, self.symbol, self.volume, self.price)
        if res is None:
            raise OrderError(f'Failed to calculate margin for {self.symbol} due to {self.mt5.error.description}')
        return res

    async def calc_profit(self) -> float:
        """Return profit in the account currency for a specified trading operation.

        Returns:
            float: Returns float value if successful

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_calc_profit(self.type, self.symbol, self.volume, self.price, self.tp)
        if res is None:
            raise OrderError(f'Failed to calculate profit for {self.symbol} due to {self.mt5.error.description}')
        return res


"""Trader class module. Handles the creation of an order and the placing of trades"""
from abc import ABC, abstractmethod
from datetime import datetime
from typing import TypeVar
from logging import getLogger
from zoneinfo import ZoneInfo

from .order import Order
from .symbol import Symbol as _Symbol
from .ticks import Tick
from .ram import RAM
from .core.models import OrderType, OrderSendResult
from .core.config import Config
from .utils import dict_to_string
from .result import Result

logger = getLogger(__name__)
Symbol = TypeVar("Symbol", bound=_Symbol)


class Trader(ABC):
    """Base class for creating a Trader object. Handles the creation of an order and the placing of trades.

    Attributes:
        symbol (Symbol): The financial instrument.
        ram (RAM): RAM instance
        order (Order): Trade order

    Class Attributes:
        config (Config): Config instance.
    """
    config: Config

    def __init__(self, *, symbol: Symbol, ram: RAM = None):
        """Initializes the order object and RAM instance

        Args:
            symbol (Symbol): Financial instrument
            ram (RAM): Risk Assessment and Management instance
        """
        self.config = Config()
        self.symbol = symbol
        self.order = Order(symbol=symbol.name)
        self.ram = ram or RAM()
        self.parameters = {}

    def set_order_limits(self, *, pips: float, tick: Tick):
        """Sets the stop loss and take profit for the order. This method uses pips as defined for forex instruments.

        Args:
            pips: Target pips
            tick: Tick object
        """
        pips = pips * self.symbol.pip
        sl, tp = pips, pips * self.ram.risk_to_reward
        if self.order.type == OrderType.BUY:
            self.order.sl, self.order.tp = round(tick.ask - sl, self.symbol.digits), round(tick.ask + tp,
                                                                                           self.symbol.digits)
            self.order.price = tick.ask
        elif self.order.type == OrderType.SELL:
            self.order.sl, self.order.tp = round(tick.bid + sl, self.symbol.digits), round(tick.bid - tp,
                                                                                           self.symbol.digits)
            self.order.price = tick.bid
        else:
            raise ValueError(f"Invalid order type: {self.order.type}")

    def set_trade_stop_levels(self, *, points: float, tick: Tick):
        """Set the stop loss and take profit levels of the order based on the points and price tick.

        Args:
            points: Target points
            tick: Tick object
        """
        points = points * self.symbol.point
        sl, tp = points, points * self.ram.risk_to_reward
        if self.order.type == OrderType.BUY:
            self.order.sl, self.order.tp = round(tick.ask - sl, self.symbol.digits), round(tick.ask + tp,
                                                                                           self.symbol.digits)
            self.order.price = tick.ask
        else:
            self.order.sl, self.order.tp = round(tick.bid + sl, self.symbol.digits), round(tick.bid - tp,
                                                                                           self.symbol.digits)
            self.order.price = tick.bid

    async def check_order(self) -> bool:
        """Check order before sending it to the broker.

        Returns:
            bool: True if order can go through else false
        """
        check = await self.order.check()
        if check.retcode != 0:
            logger.warning(f"""Invalid order for {self.symbol}
            \r\r{dict_to_string(check.request._asdict() | check.get_dict(include={'comment', 'retcode'}))}""")
            return False
        return True

    async def send_order(self) -> OrderSendResult:
        """Send the order to the broker."""
        result = await self.order.send()
        if result.retcode != 10009:
            logger.warning(f"""Unable to place order for {self.symbol}
            \r\r{dict_to_string(result.request._asdict() | result.get_dict(include={'comment', 'retcode'}))}\n""")
            return result
        logger.info(f"""Placed Trade for {self.symbol}
        \r\r{dict_to_string(result.get_dict(exclude={'request', 'retcode_external', 'retcode', 'request_id'}), multi=True)}\n""")
        await self.record_trade(result, parameters=self.parameters.copy())
        return result

    async def record_trade(self, result: OrderSendResult, parameters: dict = None, name: str = ''):
        """Record the trade in a csv file.

        Args:
            result (OrderSendResult): Result of the order send
            parameters: parameters of the trading strategy used to place the trade
            name: Name of the trading strategy
        """
        if result.retcode != 10009 or not self.config.record_trades:
            return
        params = parameters or self.parameters.copy()
        profit = await self.order.calc_profit()
        params["expected_profit"] = profit
        date = datetime.utcnow()
        date = date.replace(tzinfo=ZoneInfo("UTC"))
        params["date"] = str(date.date())
        params["time"] = str(date.time())
        res = Result(result=result, parameters=params, name=name)
        self.config.task_queue.add_task(res.to_csv)

    @abstractmethod
    async def place_trade(self, *args, **kwargs):
        """Places a trade based on the order_type."""


from dataclasses import dataclass
from typing import Literal

from ...core.constants import OrderType


@dataclass
class Tracker:
    """Keeps track of a strategy's data and state"""
    trend: Literal["ranging", "bullish", "bearish"] = "ranging"
    bullish: bool = False
    bearish: bool = False
    ranging: bool = True
    snooze: float = 0
    trend_time: float = 0
    entry_time: float = 0
    new: bool = True
    order_type: OrderType = None

    def update(self, **kwargs):
        fields = self.__dict__
        for key in kwargs:
            if key in fields:
                setattr(self, key, kwargs[key])
        if 'trend' in kwargs:
            match self.trend:
                case "ranging":
                    self.ranging = True
                    self.bullish = self.bearish = False
                case "bullish":
                    self.ranging = self.bearish = False
                    self.bullish = True
                case "bearish":
                    self.ranging = self.bullish = False
                    self.bearish = True